#!/usr/bin python  
import os
import getopt, sys  
import fasta
import array
import math
import random

# Set this to path to LIBSVM directory
#libsvmpath = "/filer/tmp/pkuksa/bin/libsvm-2.88/"
libsvmpath = "/res/seqam/pkuksa/filer/pkuksa/bin/libsvm-2.88/"
resultspath = "results/"
bindir = "bin/"

def getFeatureMap(kernelParam1=5,kernelParam2=1,type="double",Alphabet=['A','C','G','T']):
   #lists features defined by kernel of a given type
   abetSize = len(Alphabet)
   if type=="double":
      d = kernelParam1
      nKmer = abetSize**2
      f = open(resultspath + "featuremap.txt","w")
      for d1 in range(d):
        s1 = ""
        for i in range(d1):
          s1 += "_"
        for a1 in Alphabet:
          for a2 in Alphabet:
            f.write(a1 + s1 + a2 + "\n")
      f.close()
   if type=="triple":
      d = kernelParam1
      nKmer = abetSize**2
      f = open(resultspath + "featuremap.txt","w")
      for d1 in range(d):
        s1 = ""
        for i in range(d1):
          s1 += "_"
        for d2 in range(d):
          s2 = ""
          for i in range(d2):
            s2 += "_"
          for a1 in Alphabet:
            for a2 in Alphabet:
              for a3 in Alphabet:
                 f.write(a1 + s1 + a2 + s2 + a3 + "\n")
      f.close()
                 

def spectrum(sequence, kernelParam1=5, kernelParam2=1, abetSize=4, type="double"):
    # Computes feature map for a given sequence
    # Only 'double' and 'triple' features are supported now
    if type=="double":
        d = kernelParam1
        nKmer = abetSize**2
	feature = array.array('i',[0 for i in range(nKmer)])
	seq_len = len(sequence)
        for i in range(seq_len-d):
           feature[abetSize*sequence[i] + sequence[i+d]]+=1
	
    if type=="triple":
      d1 = kernelParam1
      d2 = kernelParam2
      nKmer = abetSize**3
      feature = array.array('i',[0 for i in range(nKmer)])
      seq_len = len(sequence)
      w2 = abetSize**2
      w1 = abetSize
      for i in range(seq_len-d1-d2):
        feature[w2*sequence[i] + w1*sequence[i+d1] + sequence[i+d1+d2]]+=1
    """
    if type=="mismatch":
      k = kernelParam1
      m = kernelParam2
      nKmer = abetSize**k;
      feature = array.array('i',[0 for i in range(nKmer)])
      seq_len = len(sequence)
      w = array.array('i',[0 for i in range(k)])
      for i in range(k):
        w[i] = abetSize**(k-i-1)
      for i in range(seq_len-k+1):
         h = 0
         for j in range(k):
           h+=w[j]*sequence[i+j]
         features[h]+=1
    """         

    return feature

def getLabels(dna, LabelSet):
  # Gets class labels for data for a given label set
  Label = []
  Id = []
  for item in dna:
    name = item.name.split("|")
    Label.append(name[2])
    Id.append(name[0][1:])
  LabelMap = dict((list(LabelSet)[i],i) for i in range(0,len(LabelSet)));
  LabelNum = [LabelMap[base] for base in Label]
  return LabelNum, Label, Id
	
def extractLabels(dna):
	# Extracts class labels 
        # headers are assumed to have three fields separated by "|"
        # 3rd field assumed to be the class labels
	Label = []
	Id = []
	LabelSet = set('')
	for item in dna:
		name = item.name.split("|")
		Id.append(name[0][1:])
		Label.append(name[2])
		LabelSet.add(name[2])

	LabelMap = dict((list(LabelSet)[i],i) for i in range(0,len(LabelSet)))
	LabelNum = [LabelMap[base] for base in Label]
		
	return LabelNum, list(LabelSet), Label, Id
	

def dumpSeq(dna, seqfile="input_integer.dna"):
    # Saves sequences into a file
    f = open(seqfile, 'w');
    for item in dna:
       s = fasta.numeric(fasta.cleanup2(item.sequence.upper()));
       n = len(s)
       for i in range(n-1):
         f.write(str(s[i]) + " ")
       f.write(str(s[len(s)-1]))
       f.write("\n")
    f.close()
	
def computeKernel(dna,kernelParam1=5,kernelParam2=1,abetSize=4,type="double"):
    # Computes kernel of a given type
    seqfile = resultspath + "input_integer.dna"
    dumpSeq(dna, seqfile);
    if type == "double":
	d = kernelParam1;
	kernelcmd = "doubleKernel " + seqfile + " " + str(d) + " " + str(abetSize);
    if type == "triple":
        d = kernelParam1;
	kernelcmd = "tripleKernel " + seqfile + " " + str(d) + " " + str(abetSize);
    if type == "mismatch":
        k = kernelParam1;
        m = kernelParam2;
	kernelcmd = "mismatchKernel " + seqfile + " " + str(k) + " " + str(m) + " " + str(abetSize);
    if type == "wildcard":
        k = kernelParam1;
        w = kernelParam2;
	kernelcmd = "wildcardKernel " + seqfile + " " + str(k) + " " + str(w) + " " + str(abetSize);
    if type == "gapped":
        k = kernelParam1;
        g = kernelParam2;
	kernelcmd = "gappedKernelImplicit " + seqfile + " " + str(k) + " " + str(g) + " " + str(abetSize);
              
    print(kernelcmd) 
    os.system(bindir + kernelcmd)
	
def computeKernelVector(dna,dnaquery,kernelParam1=5,kernelParam2=1,abetSize=4,type="double"):
    # Evaluates kernel of a given type between query and train data
    seqfile = resultspath + "input_integer.dna"
    queryfile = resultspath + "query_integer.dna"
    dumpSeq(dna, seqfile);
    dumpSeq(dnaquery, queryfile); 
    if type == "double":
	d = kernelParam1;
	kernelcmd = "doubleKernelVector " + seqfile + " " + queryfile + " " + str(d) + " " + str(abetSize);
    if type == "triple":
        d = kernelParam1;
	kernelcmd = "tripleKernelVector " + seqfile + " " + queryfile + " " + str(d) + " " + str(abetSize);
    if type == "mismatch":
        k = kernelParam1;
        m = kernelParam2;
	kernelcmd = "mismatchKernelVector " + seqfile + " " + queryfile + " " + str(k) + " " + str(m) + " " + str(abetSize);
    if type == "wildcard":
        k = kernelParam1;
        w = kernelParam2;
	kernelcmd = "wildcardKernelVector " + seqfile + " " + queryfile + " " + str(k) + " " + str(w) + " " + str(abetSize);
    if type == "gapped":
        k = kernelParam1;
        g = kernelParam2;
	kernelcmd = "gappedKernelVector " + seqfile + " " + queryfile + " " + str(k) + " " + str(g) + " " + str(abetSize);
              
    os.system(bindir + kernelcmd)

	
def computeFeatures(dna,kernelParam1=5,kernelParam2=1,abetSize=4,type="double"):
    # Computes features for a give kernel
    # only 'double' and 'triple' features are supported now
    featurefilename = "features.txt"	
    f = open(resultspath + featurefilename, "w")
    if type == "double":
        maxDist = int(kernelParam1)
	nKmer = abetSize**2
	nDist = maxDist
        print "Computing features"
        getFeatureMap(maxDist)
	i = 0
	for item in dna:
            sequence = fasta.numeric(fasta.cleanup2(item.sequence.upper()))
	    sequence = array.array('i', sequence)
	    for d1 in range(1,maxDist+1):
		spec = spectrum( sequence, d1, [], abetSize, type)
                for c in spec:
                  f.write(str(c) + " ")
            f.write("\n")
	    i=i+1

    if type == "triple":
        maxDist = int(kernelParam1)
	nKmer = abetSize**3
	nDist = maxDist**2
        getFeatureMap(maxDist,[],"triple")
	i = 0
	for item in dna:
            sequence = fasta.numeric(fasta.cleanup2(item.sequence.upper()))
	    sequence = array.array('i', sequence)
	    for d1 in range(1,maxDist+1):
              for d2 in range(1,maxDist+1):
		spec = spectrum( sequence, d1, d2, abetSize, type)
                for c in spec:
                  f.write(str(c) + " ")
            f.write("\n")
	    i=i+1
    f.close()    

	
def unique(s):
  u = {}
  for i in s: 
    u[i] = 1
  return u.keys()


def crossvalind(group, k):
   # Generates k-fold crossvalidation indices
   cvidx = array.array('i',[0 for i in range(len(group))])
   classids = unique(group)   
   for c in classids:
     ind = array.array('i', [i for i in range(len(group)) if group[i]==c])
     indf = array.array('i', [0 for i in group if i==c])
     irange = array.array('i', [i+1 for i in range(len(ind))])
     for i in range(len(ind)):
        indf[i] = int(math.ceil(float(irange[i])/(float(len(ind))/k)))
     items = [j+1 for j in range(k)]
     random.shuffle(items)
     rperm = [j for j in range(len(ind))]
     random.shuffle(rperm)
     for j in range(len(ind)):
        cvidx[ind[rperm[j]]] = items[indf[j]-1]
   
   return cvidx

def saveLabels(Labels,filename="datalabels.txt"):
   f = open(filename, "w")
   for label in Labels:
      f.write(str(label) + "\n")
   f.close();
	
def usage():
	
	print "Usage: ", sys.argv[0], " [ht:K:Q:V:] dna_file"
	print '  -h or --help: print this help'
	print '  -t or --type: type of analysis to perform on sequences in dna_file:'
	print '       kernel   - compute similarity/kernel '
	print '       feature  - compute features '
	print '       svm      - cross-validation using a multiclass linear SVM'
	print '       nn       - cross-validation using nearest neighbor (NN) classifier'
        print '       predict  - compute assignments for query sequences in query_file using reference/train sequences in dna_file'
        print '  -K : similarity function/kernel type'
        print '       double'
        print '       triple'
        print '       mismatch'
        print '       gapped'
        print '  -Q : specifies query_dna file when -t predict is used'
        print '  -V : number of folds for cross-validation'
        print '  -k, -m, -w, -g, -d : kernel parameters'
        print '       k - k-mer length (mismatch, gapped)'
        print '       m - max. number of mismatches (mismatch)'
        print '       w - max. number of wildcards (wildcard)'
        print '       g - max. number of gaps (gapped)'
        print '       d - max. distance between samples (double,triple)'

def main():

        if (len(sys.argv)==1):
           usage()
           sys.exit(2)

	try:
	   opts, args = getopt.getopt(sys.argv[1:], "d:k:m:w:g:ht:vK:Q:V:", ["d=","help", "type="])
	except getopt.GetoptError, err:
		# print help information and exit:
		print str(err)
		usage()
		sys.exit(2)
        # Default parameters
	output = "kernel"
        kerneltype = "double"
        queryfilename = "test.fasta"
        kernelParam1 = 5
        kernelParam2 = 1
        nfold = 10
	verbose = False

	for o, a in opts:
		if o == "-v":
			verbose = True
		elif o in ("-h", "--help"):
			usage()
			sys.exit()
		elif o in ("-t", "--type"):
			output = a
                elif o == "-d":
                    kernelParam1 = a;
                elif o == "-k":
                    kernelParam1 = a;
                elif o == "-m":
                    kernelParam2 = a;
                elif o == "-g":
                    kernelParam2 = a;
                elif o == "-w":
                    kernelParam2 = a;
                elif o == "-K":
                    kerneltype = a;
                elif o == "-Q":
                    queryfilename = a; 
                elif o == "-V":
                    nfold = int(a); 
		else:
			assert False, "unhandled option"
	
	filename = args[0] # train_data 

	# Read DNA sequences			
	dna = fasta.read_fasta( open(filename, 'r').readlines() )	
	L, labelSet, labels, ids  = extractLabels(dna)
        NumLabelMap=dict( (i,list(labelSet)[i]) for i in range(len(labelSet)) )
        
        f = open(resultspath + "classnames.txt","w")
        for i in range(len(labelSet)):
          f.write(str(i)+"\t"+NumLabelMap[i]+"\n")
        f.close()

        if (queryfilename != "" and output == "predict"):
           dnaquery=fasta.read_fasta( open(queryfilename, 'r').readlines())
           Lquery, labelsquery, idsquery = getLabels(dnaquery, labelSet)
	
	
	if output == 'kernel':
		computeKernel(dna,kernelParam1,kernelParam2,4,kerneltype)
                os.system(bindir + "normalizekernel kernel.txt")
                os.system("mv kernel.txt kernel.txt.normalized " + resultspath)
		
	elif output == 'feature':
		# Compute features
		computeFeatures(dna,kernelParam1,kernelParam2,4,kerneltype)
		
	elif output == 'svm':
                # Runs cross-validation using svm with selected kernel
                print "Running %d-fold cross-validation using SVM"%(nfold)
                kernelfilename = "kernel.txt"
                labelsfilename = resultspath + "datalabels.txt"
		computeKernel(dna,kernelParam1,kernelParam2,4,kerneltype)
                saveLabels(L,labelsfilename)
                # Train multi-class SVM from precomputed kernel

                # Comvert kernel to libsvm format
                convertcmd = "clibsvm " + kernelfilename + " " + labelsfilename
                os.system(bindir + convertcmd)
                os.system("mv kernel.txt kernel.txt.libsvm " + resultspath)

                # Run n-fold cross-validation
	        libsvmcmd = libsvmpath + "svm-train -t 4 -c 1 -v " + str(nfold) + " " + resultspath + kernelfilename + ".libsvm"
                os.system(libsvmcmd)

        elif output == "nn":
          cvIdx = crossvalind(L, nfold)
          saveLabels(cvIdx, resultspath + "cv_idx.txt");
          print "Running %d-fold cross-validation using NN" % (nfold)
          computeKernel(dna,kernelParam1,kernelParam2,4,kerneltype)
          os.system(bindir + "normalizekernel kernel.txt")
          os.system("mv kernel.txt kernel.txt.normalized " + resultspath)
          
          totalerr = 0
          baltotalerr = 0
          baltotalerrstd = 0
          totalerrstd = 0
          balerr = array.array('f', [0 for j1 in range(nfold)])
          folderr = array.array('f', [0 for j1 in range(nfold)])
          for i in range(nfold):
            testidx = [j for j in range(len(cvIdx)) if cvIdx[j] == i+1]
            trainidx = [j for j in range(len(cvIdx)) if cvIdx[j] != i+1]
            f = open(resultspath + "kernel.txt.normalized","r")
            j = 0
            p = 0
            errcnt = 0  
            classcnt = array.array('f',[0 for j1 in range(len(labelSet))])
            balerrcnt = array.array('f',[0 for j1 in range(len(classcnt))])
            for line in f.readlines():
               if j in testidx:
                 values = line.split(" ")
                 max_val = 0
                 max_ind = -1 
                 for j1 in trainidx:
		    if float(values[j1]) > max_val:
                       max_val = float(values[j1])
                       max_ind = j1
                 if (L[max_ind] != L[j]):
                   errcnt+=1 # 0-1 loss
                   balerrcnt[L[j]]+=1 # balanced loss
                 classcnt[L[j]]+=1
                 p+=1
               j+=1
            f.close()   
            folderr[i] = float(errcnt)/p
            totalerr += float(errcnt)/p
            classestest = [j1 for j1 in range(len(labelSet)) if classcnt[j1]>0]
            for j1 in classestest:
              balerrcnt[j1] /= classcnt[j1]
            balerr[i] = sum(balerrcnt)/len(classestest)
            baltotalerr += balerr[i]
            print "Fold %d: %d/%d (%.4f, %.4f)" % (i,errcnt,p,float(errcnt)/p,balerr[i])
             
          for j1 in range(nfold):
            baltotalerrstd += (balerr[j1] - baltotalerr/nfold)**2
            totalerrstd += (folderr[j1] - totalerr/nfold)**2
          print "Cross-validation accuracy = %.4f%%+-%.4f%%" % ((1-totalerr/nfold)*100, math.sqrt(totalerrstd/(nfold-1))*100)
          print "Cross-validation balanced accuracy = %.4f%%+-%.4f%%" % ((1-baltotalerr/nfold)*100, math.sqrt(baltotalerrstd/(nfold-1))*100)


        elif output == "predict":
                # Trains SVM model on train_data and tests on test_data

                # Train SVM
                kernelfilename = resultspath + "kernel.train"
                trainkernelfilename = resultspath + "kernel.train"
                labelsfilename = resultspath + "labels.train"
                modelfilename =  resultspath + "model.libsvm"
		computeKernel(dna,kernelParam1,kernelParam2,4,kerneltype)
                os.system("mv kernel.txt " + kernelfilename)
                f = open(labelsfilename,"w")
                for label in L:
                  f.write(str(label) + "\n")
                f.close();
                convertcmd = "clibsvm " + kernelfilename + " " + labelsfilename
                os.system(bindir + convertcmd)
	        libsvmcmd = libsvmpath + "svm-train -t 4 " + kernelfilename + ".libsvm " + modelfilename
                os.system(libsvmcmd)
                print "Finished SVM training"
                # Test SVM
                kernelfilename = resultspath + "kernel.test"
                querylabelsfilename = resultspath + "labels.test";
                testfilename =  resultspath + "kernel.test.libsvm"
                outputfilename = resultspath + "predict.libsvm"
                print "Testing on query file"
		computeKernelVector(dna,dnaquery,kernelParam1,kernelParam2,4,kerneltype)
                os.system("mv kernel.txt " + kernelfilename)
                saveLabels(Lquery, querylabelsfilename)
                convertcmd = "clibsvm " + kernelfilename + " " + querylabelsfilename
                os.system(bindir + convertcmd)
                
	        libsvmpredictcmd = libsvmpath + "svm-predict " + testfilename + " " + modelfilename + " " + outputfilename
                os.system(libsvmpredictcmd)

                f = open(outputfilename, "r")
                predClassInt = []
                predClass = []
                for line in f.readlines():
                  predClassInt.append(int(line))
                  predClass.append(NumLabelMap[int(line)])
                f.close()
                f = open(resultspath + "predictions.txt","w")
                for i in range(len(predClass)):
                   f.write(str(predClassInt[i])+", "+predClass[i]+"\n")
                f.close()

                f = open(trainkernelfilename, "r")
                diag_train = array.array('f',[0 for i in range(len(dna))]) 
                i = 0;
                for line in f.readlines():
                   values = line.split(" ")
                   diag_train[i] = float(values[i])
                   i+=1
                f.close()
		computeKernel(dnaquery,kernelParam1,kernelParam2,4,kerneltype)
                querykernelfilename = resultspath + "kernel.query"
                os.system("mv kernel.txt " + querykernelfilename)
                f = open(querykernelfilename, "r")
                diag_query = array.array('f',[0 for i in range(0,len(dnaquery))]) 
                i = 0
                for line in f.readlines():
                   values = line.split(" ")
                   diag_query[i] = float(values[i])
                   i+=1
                f.close()
                f = open(kernelfilename, "r")
                fout = open(resultspath + "ranking.txt", "w")
                fout2 = open(kernelfilename + ".normalized","w")
                j = 0 
                nn_error = 0
                for line in f.readlines():
                  values = line.split(" ")
                  max_ind = 1
                  max_value = 0
                  for i in range(len(values)):
                    val = float(values[i]) / math.sqrt(diag_query[j]) / math.sqrt(diag_train[i])
                    fout2.write(str(val)+" ") 
                    if val > max_value:
                      max_value = val
                      max_ind = i+1
                  #fout.write('%f %d %s true: %s\n' % (max_value, max_ind, NumLabelMap[L[max_ind]], NumLabelMap[Lquery[j]]))
                  fout.write('%s\t%s\t%f\n' % (dnaquery[j].name[1:], dna[max_ind].name[1:], max_value))
                  if (L[max_ind] != Lquery[j]):
                     nn_error += 1
                  fout2.write("\n")
                  j+=1
                f.close()
                fout.close() 
                fout2.close()
                print "NN accuracy = %.4f%% (%d/%d)\n" % (100*(1.0-float(nn_error)/j),j-nn_error,j)
	
if __name__ == "__main__":
	main()
