#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shared.h"

#define MIN_REQUIRED 2

int help()
{
 printf("Usage: clibsvm  <Kernel-file> <Label-file>\n");
 printf("\t Kernel-file : text file with Gram kernel matrix\n");
 printf("\t Label-file : text file with labels (one label (integer) per line)\n");
 return 1;
}

int main(int argc, char **argv)
{  int nStr, nl;
   int i, j;
   double **K;
   int **lab, *len, *len1;
   char *kernelfilename, *labelfilename;
   FILE *libsvmfile;

   if (argc != MIN_REQUIRED+1)
     return help();

   kernelfilename = argv[1];
   labelfilename = argv[2];

   len = (int *)malloc(MAXNSTR*sizeof(int));
   len1 = (int *)malloc(MAXNSTR*sizeof(int));
   printf("Reading %s\n",kernelfilename);
   K = loadStringsDouble(kernelfilename, len, &nStr);
   printf("Read %d strings\n", nStr);
   lab = loadStrings(labelfilename, len1, &nl);   
   printf("Read %d labels\n", nl);

   if (nStr != nl)
   {
     printf("ERROR: number of labels and kernel matrix size do not match");
     exit(1);
   }
   libsvmfile = fopen(strcat(kernelfilename,".libsvm"),"w");
   if (!libsvmfile)
   {
     perror("Cannot write output");
     exit(1);
   } 
   for (i = 0; i < nStr; ++i)
   { 
     fprintf(libsvmfile,"%d 0:%d ", lab[i][0], i+1);
     for (j = 0; j < len[i]-1; ++j)
        fprintf(libsvmfile,"%d:%.4f ", j+1, K[i][j]);
     fprintf(libsvmfile,"%d:%.4f\n", len[i], K[i][len[i]-1]);
   }
   fclose(libsvmfile);   

   return 0;
}

