#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shared.h"

#define ARG_REQUIRED 4

Features *extractDouble(int **S, int *len, int nStr, int d1)
{
  int i, j;
  int n, nfeat;
  int *group;
  int *features;
  int *s;
  int c;
  Features *F;

  nfeat = 0;
  for (i = 0; i < nStr; ++i)
    nfeat += len[i] - d1;
  group = (int *)malloc(nfeat*sizeof(int));
  features = (int *)malloc(nfeat*2*sizeof(int *));
  c = 0;
  for (i = 0; i < nStr; ++i)
  {
    n = len[i] - d1;
    s = S[i];
    for (j = 0; j < n; ++j)
    {
      features[c] = s[j];
      features[c+nfeat] = s[j+d1];
      group[c] = i;
      c++;
    }
  }
  if (nfeat!=c)
    printf("Something is wrong...\n");
  F = (Features *)malloc(sizeof(Features));
  (*F).features = features;
  (*F).group = group;
  (*F).n = nfeat;
  return F;
}

int help()
{
  printf("Usage: doubleKernel <Sequence-file> <d> <Alphabet-size>\n");
  printf("\t Sequence-file : file with sequence data\n");
  printf("\t d : kernel parameter (max. distance between samples), d>0\n");
  printf("\t Alphabet-size: size of the sequence alphabet, >0\n");
  printf("\t IMPORTANT: sequence elements are assumed to be\n\tin the range [0,AlphabetSize - 1]\n");
  return 1;
}

char isVerbose;

int *doubleKernelVector(int **S, int d, int nStr, int na, int *len, int nq);

int main (int argc, char **argv)
{
  char *databasefilename, *queryfilename;
  int d, na, nStr, nq;
  int i, j;
  int *K, *K1;
  int **S, **Sq, **Sd, *len, *lenq, *lend;
  char *kernelfilename;

  isVerbose = 0;
  if (argc != ARG_REQUIRED+1)
  {
    return help();
  }

  databasefilename = argv[1];
  queryfilename = argv[2];
  d = atoi(argv[3]);
  na = atoi(argv[4]);
  if (d <= 0 || na <=0)
  {
    printf("Error in input parameters\n");
    return help();
  }
  lend = (int *)malloc(MAXNSTR*sizeof(int));
  printf("Reading %s\n",databasefilename);
  Sd = loadStrings(databasefilename, lend, &nStr);
  printf("Read %d strings\n", nStr);
  lenq = (int *)malloc(MAXNSTR*sizeof(int));
  printf("Reading %s\n", queryfilename);
  Sq = loadStrings(queryfilename, lenq, &nq);
  printf("Read %d strings\n", nq);
  S = (int **)malloc((nStr+nq)*sizeof(int *));
  len = (int *)malloc((nq + nStr)*sizeof(int));
  for (i = 0; i < nq; ++i)
  {
      S[i] = (int *)malloc(lenq[i]*sizeof(int));
      len[i] = lenq[i];
      for (j = 0; j < lenq[i]; ++j) 
        S[i][j] = Sq[i][j];
  }
  for (i = 0; i < nStr; ++i)
  {
      S[i+nq] = (int *)malloc(lend[i]*sizeof(int));
      len[i+nq] = lend[i];
      for (j = 0; j < lend[i]; ++j) 
        S[i+nq][j] = Sd[i][j];
  }

  for (i = 0; i < nStr; ++i)
     free(Sd[i]);
  free(Sd); 
  for (i = 0; i < nq; ++i)
     free(Sq[i]);
  free(Sq);  

  K = doubleKernelVector(S, d, nStr+nq, na, len, nq);
  K1 = (int *)malloc(nq*nStr*sizeof(int));
  for (i = 0; i < nq; ++i)
  {
    for (j = nq; j < nStr+nq; ++j)
       K1[i + (j-nq)*nq] = K[i + j*(nStr+nq)];
  }
  free(K);
  kernelfilename = "kernel.txt";
  saveKernelInt(K1, nq, nStr,"%d ",kernelfilename);

  return 0;
}

int *doubleKernelVector(int **S, int d, int nStr, int na, int *len, int nq)
{
  int i, d1, maxIdx, nfeat;
  int *K;
  int *features_srt, *group_srt, *sortIdx;
  Features *features;

  K = (int *)malloc(nStr*nStr*sizeof(int));
  for (i=0; i<nStr*nStr; ++i)
    K[i]=0;
  printf("Computing double(1,%d) kernel\n", d);
  for (d1 = 1; d1 <= d; ++d1)
  {
      if ( isVerbose ) printf("Extracting features..."), fflush( stdout );
      features = extractDouble(S,len,nStr,d1);
      nfeat = (*features).n;
      printf("d=%d: %d features\n", d1, nfeat);
      maxIdx = 0;
      for (i = 0; i < nfeat*2; ++i)
        maxIdx = maxIdx > (*features).features[i] ? maxIdx : (*features).features[i];
      if (na < maxIdx+1)
      {
        printf("WARNING: Sequence elements are outside the specified range [0,%d]\n",na);
        printf("\tUsing [0,%d] instead\n", maxIdx);
        na = maxIdx+1;
      }
      if (isVerbose)
      {
        printf("done.\n");
        printf("Sorting...");
        fflush( stdout );
      }
      sortIdx = cntsrtna((*features).features,2,(*features).n,na);
      if (isVerbose)  printf("done.\n");
      features_srt = (int *)malloc(nfeat*2*sizeof(int *));
      group_srt = (int *)malloc(nfeat*sizeof(int));
      for (i = 0; i < nfeat; ++i)
      {
        features_srt[i]=(*features).features[sortIdx[i]];
        features_srt[i+nfeat]=(*features).features[sortIdx[i]+nfeat];
        group_srt[i] = (*features).group[sortIdx[i]];
      }
      free(sortIdx);
      free((*features).features);
      free((*features).group);
      free(features);
      if (isVerbose)
      {
        printf("Counting...");
        fflush( stdout );
      }
      countAndUpdate_vector(K,features_srt,group_srt,2,nfeat,nStr,nq);
      free(features_srt);
      free(group_srt);
      if (isVerbose)
      {
        printf("done.");
      }
  }
  
  return K;
}
