#include "shared.h"
#include <stdlib.h>
#include <stdio.h>

#define ARG_REQUIRED 4 

Features *extractKmers(int **S, int *len, int nStr, int k)
{
  int i, j, jj;
  int n, nfeat;
  int *group;
  int *features;
  int *s;
  int c;
  Features *F;

  nfeat = 0;
  for (i = 0; i < nStr; ++i)
    nfeat += len[i] - k + 1;
  group = (int *)malloc(nfeat*sizeof(int));
  features = (int *)malloc(nfeat*k*sizeof(int *));
  c = 0;
  for (i = 0; i < nStr; ++i)
  {
    n = len[i] - k + 1;
    s = S[i];
    for (j = 0; j < n; ++j)
    {
      for (jj = 0; jj < k; ++jj)
        features[c + jj*nfeat] = s[j+jj];
      group[c] = i;
      c++;
    }
  }
  if (nfeat!=c)
    printf("Something is wrong...\n");
  F = (Features *)malloc(sizeof(Features));
  (*F).features = features;
  (*F).group = group;
  (*F).n = nfeat;
  return F;
}


int help()
{
 printf("Usage: gappedKernel <Sequence-file> <k> <g> <Alphabet-size>\n");
 printf("\t Sequence-file : file with sequence data\n");
 printf("\t k : kernel parameter (k-mer length), k>0\n");
 printf("\t g : kernel parameter (max. number of gaps), g>=0\n");
 printf("\t\t i.e. \"g-mer\" length is g+k\n");
 printf("\t Alphabet-size: size of the sequence alphabet, >0\n");
 printf("\t IMPORTANT: sequence elements are assumed to be\n\tin the range [0,AlphabetSize - 1]\n");
  return 1;
}

char isVerbose;
double *gappedKernel(Features *features, int k, int g, int nStr, int na, int *len);
int main (int argc, char **argv)
{
  int k, g, na;
  char *filename;
  int i;
  int *len;
  int **S;
  double *K;
  int nStr, nfeat, ngmers, flen, glen;
  Features *features;
  int maxIdx;
  char *kernelfilename;

  isVerbose = 0;
  if (argc != ARG_REQUIRED+1)
  {
    return help();
  }

  filename = argv[1];
  k = atoi( argv[2] );
  g = atoi( argv[3] );
  na = atoi( argv[4] );

  if ( !(k > 0 && g > 0 && na > 0) )
  {
    printf("Error in input parameters\n");
    return help();
  }

    glen = g + k;
    flen = k;
    len = (int *)malloc(MAXNSTR*sizeof(int));
    printf("Reading %s\n",filename);
    S = loadStrings(filename, len, &nStr);
    printf("Read %d strings\n", nStr);
    features = extractKmers(S,len,nStr,glen);
    nfeat = (*features).n;
    ngmers = (*features).n;
    printf("Extracted %d (k+g)-mers\n", ngmers);
    fflush( stdout );
    for (i = 0; i < nStr; ++i)
      free(S[i]);
    free(S); 
    /* check sequence elements range */
     maxIdx = 0;
     for (i = 0; i < nfeat*k; ++i)
       if (maxIdx < (*features).features[i])
         maxIdx = (*features).features[i];
     if (na < maxIdx + 1)
     {
       printf("WARNING: sequence elements are outside of the specified range [0,%d]\n",na-1);
       printf("\t using [0 %d] instead\n", maxIdx);
       na = maxIdx + 1;
     }
    K = gappedKernel(features, k, g, nStr, na, len);
    kernelfilename = "kernel.txt";
    saveKernelDouble(K,nStr,nStr,"%.0f ",kernelfilename);
    free(K);
     
    return 0;
}

double *gappedKernel(Features *features, int k, int g, int nStr, int na, int *len)
{
    double *K;
    int i, glen, flen, nfeat, nComb, ngmers;
    int *sortIdx;
    Combination *combinations;

    glen = g + k;
    flen = k;
    combinations = getCombinations(glen, flen);
    ngmers = (*features).n;
    K = (double *)malloc(nStr*nStr*sizeof(double));
    for (i=0; i<nStr*nStr; ++i)
	K[i]=0;
    nComb = (int)nchoosek(glen, flen);
    nfeat = ngmers * nComb;
    printf("Computing (k=%d,g=%d) gapped kernel (%d feat.) \n",k,g,nfeat);
    if (isVerbose) printf("flen=%d, ngmers=%d, nfeat=%d, ncomb=%d\n",k,nfeat,ngmers,nComb);
    if (isVerbose) printf("Sorting..."), fflush( stdout );
    sortIdx = cntsrtna_gapped((*features).features,glen,flen,ngmers,na,combinations);
    if (isVerbose) printf("Counting...");
   countAndUpdate_double_gapped(K,(*features).features,(*features).group,sortIdx,flen,ngmers,nStr,combinations);
    
    return K;
}
