#include "shared.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>


#define ARG_REQUIRED 4 

Features *extractKmers(int **S, int *len, int nStr, int k)
{
  int i, j, jj;
  int n, nfeat;
  int *group;
  int *features;
  int *s;
  int c;
  Features *F;

  nfeat = 0;
  for (i = 0; i < nStr; ++i)
    nfeat += len[i] - k + 1;
  group = (int *)malloc(nfeat*sizeof(int));
  features = (int *)malloc(nfeat*k*sizeof(int *));
  c = 0;
  for (i = 0; i < nStr; ++i)
  {
    n = len[i] - k + 1;
    s = S[i];
    for (j = 0; j < n; ++j)
    {
      for (jj = 0; jj < k; ++jj)
        features[c + jj*nfeat] = s[j+jj];
      group[c] = i;
      c++;
    }
  }
  if (nfeat!=c)
    printf("Something is wrong...\n");
  F = (Features *)malloc(sizeof(Features));
  (*F).features = features;
  (*F).group = group;
  (*F).n = nfeat;
  return F;
}

double power(int base, int exp)
{
   int i;
   double p;
   p = 1.0;
   for (i = 0;  i < exp; ++i)
     p *= (double)base;
   
   return p;
}


int help()
{
 printf("Usage: mismatchKernel <Sequence-file> <k> <m> <Alphabet-size>\n");
 printf("\t Sequence-file : file with sequence data\n");
 printf("\t k : kernel parameter (k-mer length), k>0\n");
 printf("\t m : kernel parameter (max. number of mismatches), m>=0\n");
 printf("\t Alphabet-size: size of the sequence alphabet, >0\n");
 printf("\t IMPORTANT: sequence elements are assumed to be\n\tin the range [0,AlphabetSize - 1]\n");
  return 1;
}

char isVerbose;
double *mismatchKernel(Features *features, int k, int m, int nStr, int na, int *len);

int main (int argc, char **argv)
{
  int k, m, na;
  char *filename;
  char *kernelfilename;
  int i;
  int *len;
  int **S;
  double *K;
  int nStr, nfeat, maxIdx;
  Features *features;

  isVerbose = 0;

  if (argc != ARG_REQUIRED+1)
  {
    return help();
  }

  filename = argv[1];   /* file with sequences */
  k = atoi( argv[2] );  /* k-mer length */
  m = atoi( argv[3] );  /* number of mismatches */
  na = atoi( argv[4] ); /* alphabet size */

  if (m >= 0 && k > 0 && na > 0 && m < k)
  {
    
    /* read strings */
    /* use line feeds to end lines */
    len = (int *)malloc(MAXNSTR*sizeof(int));
    printf("Reading %s\n",filename);
    S = loadStrings(filename, len, &nStr);
    printf("Read %d strings\n", nStr);
   
     features = extractKmers(S,len,nStr,k);
     nfeat = (*features).n;
     printf("Extracted %d features\n", nfeat);
     fflush( stdout );
      
     for (i = 0; i < nStr; ++i)
       free(S[i]);
     free(S);
     free(len);
    /* check sequence elements range */
     maxIdx = 0;
     for (i = 0; i < nfeat*k; ++i)
       if (maxIdx < (*features).features[i])
         maxIdx = (*features).features[i];
     if (na < maxIdx + 1)
     {
       printf("WARNING: sequence elements are outside of the specified range [0,%d]\n",na-1);
       printf("\t using [0 %d] instead\n", maxIdx);
       na = maxIdx + 1;
     }
  
    K = mismatchKernel(features, k, m, nStr, na, len);
    kernelfilename = "kernel.txt";
    saveKernelDouble(K,nStr,nStr,"%.0f ",kernelfilename);
    free(K);
    return 0;
  }
  else
  {
    printf("Error in input parameters\n");
    return help();
  }
}


double *mismatchKernel(Features *features, int k, int m, int nStr, int na, int *len)
{
    int **Ks;
    double *K;
    double *w;
    int *comb1;
    int *sortIdx;
    int i, j, ii, jj, nfeat, flen;
    int nComb;
    int num_max_mismatches, mmax;
    int nkj;
    
    Combination **combinations;

    nfeat = (*features).n;
    mmax = 2*m < k-1 ? 2*m : (k-1);
    combinations = (Combination **)malloc((mmax+1)*sizeof(Combination *));
    for (i = 0; i <= mmax; ++i)
          combinations[i] = getCombinations(k, k-i);

    /* create array of kernels */
    Ks = (int **)malloc((k+1)*sizeof(int *)); 
    for (i = 0; i <= mmax; ++i)
    {
      Ks[i] = (int *)malloc(nStr*nStr*sizeof(int));
      for (j = 0; j < nStr*nStr; ++j)
        Ks[i][j] = 0;
    }

    /* VERY IMPORTANT: intersection sizes*/
    /* NOTE: currently up to m=3 only */
    /* w[i] - int. size for distance i */
    w=(double *)malloc((k+1)*sizeof(double));
    w[0] = 1;
    for (i=1; i <= m; ++i)
      w[0] = w[0] + nchoosek(k, i)*power(na-1, i);

    if (m == 1)
    {
      w[1] = na;
      w[2] = 2;
    } 

    if (m==2)
    {
      w[1] = 1 + k * (na-1) + (k-1) * (double)(na-1)*(na-1);
      w[2] = 1 + 2*(k-1)*(na-1) + (double)(na-1)*(na-1);
      w[3] = 6*(na-1);
      w[4] = 6;
    }
    if (m==3)
    {
      w[1]=1+k*(na-1) + nchoosek(k,2)*(na-1)*(na-1)+ nchoosek(k-1,2)*(na-1)*(na-1)*(na-1);
      w[2]=1+k*(na-1)+(nchoosek(k,2)+ nchoosek(k-2,2))*(na-1)*(na-1)+(double)(k-2)*(na-1)*(na-1)*(na-1);
      w[3]=1+3*(na-1)+(6*(k-3)+3)*(na-1)*(na-1)+(double)(na-1)*(na-1)*(na-1);
      w[4]=2+nchoosek(k-1,2)*(na-1)+12*(double)(na-1)*(na-1);
      w[5]=9+10*(na-1)+(double)(na-1)*(na-1);
      w[6]=nchoosek(6,3);
    }

    for (i = 0; i<=(2*m < k ? 2*m : k); ++i)
       printf("w[%d] = %.0f ", i, w[i]);
    printf("\n");

     printf("Computing (k=%d,m=%d) mismatch kernel\n", k, m);
     for (i = 0; i <= mmax; ++i)
     {
       nComb = (int)nchoosek(k, k-i);
       printf("d=%d (%d iteration(s))\n",i,(int)combinations[i][0].num_comb);
       for (ii = 0; ii < nComb; ++ii)
       {
         flen = k-i; /* feature length */
         comb1 = (int *)malloc(flen*sizeof(int));
         for (jj = 0; jj < combinations[i][0].k; ++jj)
         {
           comb1[jj] = combinations[i][0].comb[ii+jj*(int)combinations[i][0].num_comb];
           if (isVerbose) printf("%d ", comb1[jj]);
         }
         if (isVerbose) printf("\n");
         if (isVerbose) printf("Sorting..."), fflush( stdout );
         sortIdx = cntsrtna_comb((*features).features,flen,nfeat,na,comb1);
         if (isVerbose) printf("Counting..."), fflush( stdout );
         countAndUpdate_comb(Ks[i],(*features).features,(*features).group,flen,nfeat,nStr,sortIdx,comb1);
         free(sortIdx);
         free(comb1);
       }
     }
     num_max_mismatches = 2*m < k ? 2*m : k;
     for (i = 1; i<=(num_max_mismatches < k-1 ? num_max_mismatches:k-1);++i)
     {
        for (j = 0; j <= i-1; ++j)
        {
          nkj = (int)nchoosek(k-j, (i-j));
          for (jj = 0; jj < nStr*nStr; ++jj)
            Ks[i][jj] = Ks[i][jj] - nkj*Ks[j][jj];
        }
     } 
     if (num_max_mismatches == k)
     {
       Ks[k] = (int *)malloc(nStr*nStr*sizeof(int));
       for (j = 0; j < nStr; ++j)
         for (jj = 0; jj < nStr; ++jj)
            Ks[k][j+jj*nStr] = (len[j]-k+1)*(len[jj]-k+1);
       for (i = 0; i <= k-1; ++i)
         for (jj = 0; jj < nStr*nStr; ++jj)
            Ks[k][jj] = Ks[k][jj] - Ks[i][jj];
     }
     K = (double *)malloc(nStr*nStr*sizeof(double));
     for (j = 0; j < nStr*nStr; ++j)
       K[j] = 0;
     for (i = 0; i <= num_max_mismatches; ++i)
     {
         for (j = 0; j < nStr*nStr; ++j)
            K[j] = K[j] + (double)Ks[i][j] * w[i];
     }
     for (i = 0; i <= num_max_mismatches; ++i)
       free(Ks[i]);
     free(Ks); free(w); free(combinations);
     return K;
 } 
