#include "shared.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

int **loadStrings(char *filename, int* len, int *nStr)
{
  int **S;
  char *linetemp, *line, *next_elem;
  int i, c;
  int elem;
  FILE *inpfile;

  inpfile = fopen ( filename, "r" );
  nStr[0] = MAXNSTR;
  S = (int **)malloc(nStr[0]*sizeof(int *));
  if ( inpfile )
  {
     line = (char *)malloc(STRMAXLEN*sizeof(char));
     i = 0;
     while( fgets( line, STRMAXLEN, inpfile ) )
     {
        c = 0;
        linetemp = (char *)malloc(STRMAXLEN*sizeof(char));
        strcpy(linetemp, line);
        next_elem = strtok(linetemp," ");
        if (next_elem != NULL) ++c;
        while (next_elem != NULL)
        {
          next_elem = strtok(NULL," ");
          if (next_elem != NULL) ++c;
        }
        free(linetemp);
        if (c==0) printf("ERROR: empty line detected"), exit(1);
        len[i] = c;
         
        S[i] = (int *)malloc(len[i]*sizeof(int));
        linetemp = (char *)malloc(STRMAXLEN*sizeof(char));
        strcpy(linetemp, line);
        c = 0;
        next_elem = strtok(linetemp," ");
        while (next_elem != NULL)
        {
          elem = atoi( next_elem );
          S[i][c] = elem;
          c++;
          next_elem = strtok(NULL," ");
        }
        free(linetemp);
        i++;
     }
     nStr[0] = i;
     fclose(inpfile);
     free(line);
  }
  else
  {
    perror( filename );   
    exit(1);
  }
  return S;
}


double **loadStringsDouble(char *filename, int* len, int *nStr)
{
  double **S;
  char *linetemp, *line, *next_elem;
  char **tmpptr;
  int i, c;
  double elem;
  FILE *inpfile;
  tmpptr = (char **)malloc(sizeof(char *));
  inpfile = fopen ( filename, "r" );
  nStr[0] = MAXNSTR;
  S = (double **)malloc(nStr[0]*sizeof(double *));
  if ( inpfile )
  {
     line = (char *)malloc(STRMAXLEN*sizeof(char));
     i = 0;
     while( fgets( line, STRMAXLEN, inpfile ) )
     {
        c = 0;
        linetemp = (char *)malloc(STRMAXLEN*sizeof(char));
        strcpy(linetemp, line);
        next_elem = strtok(linetemp," ");
        if (next_elem != NULL) ++c;
        while (next_elem != NULL)
        {
          next_elem = strtok(NULL," ");
          if (next_elem != NULL) ++c;
        }
        free(linetemp);
        if (c==0) printf("ERROR: empty line detected"), exit(1);
        len[i] = c;
         
        S[i] = (double *)malloc(len[i]*sizeof(double));
        linetemp = (char *)malloc(STRMAXLEN*sizeof(char));
        strcpy(linetemp, line);
        c = 0;
        next_elem = strtok(linetemp," ");
        while (next_elem != NULL)
        {
          elem = strtod(next_elem, tmpptr);
          S[i][c] = elem;
          c++;
          next_elem = strtok(NULL," ");
        }
        free(linetemp);
        i++;
     }
     nStr[0] = i;
     fclose(inpfile);
     free(line);
  }
  else
  {
    perror( filename );   
    exit(1);
  }
  return S;
}

int *cntsrtna(int *sx, int k, int r, int na)
{
    int *sxc, *bc, *sxl, *cc, *regroup;
    int i, j;
 
    sxc = (int *)malloc(na*sizeof(int));
    bc  = (int *)malloc(na*sizeof(int));
    sxl = (int *)malloc(r*sizeof(int));
    cc  = (int *)malloc(r*sizeof(int));
    regroup = (int *)malloc(r*sizeof(int));

    for (i = 0; i < r; ++i)
       regroup[i]=i;
    for (j = k-1; j >= 0; --j)
    {
       for(i = 0; i < na; ++i)
         sxc[i]=0;
       for (i = 0; i < r; ++i)
       {
                cc[i]=sx[regroup[i]+j*r];
		sxc[ cc[i] ]++;
       }
       bc[0]=0;
       for (i = 1;i < na; ++i)
           bc[i]=bc[i-1]+sxc[i-1];
       for (i = 0; i < r; ++i)
            sxl[bc[ cc[i] ]++] = regroup[i];
       for (i = 0; i < r; ++i)
            regroup[i] = sxl[i];
    }
    free(sxl); free(bc); free(sxc); free(cc);

    return regroup;
}

int *cntsrtna_comb(int *sx, int k, int r, int na, int *comb)
{
    int *sxc, *bc, *sxl, *cc, *regroup;
    int i, j;

    sxc = (int *)malloc(na*sizeof(int));
    bc  = (int *)malloc(na*sizeof(int));
    sxl = (int *)malloc(r*sizeof(int));
    cc  = (int *)malloc(r*sizeof(int));
    regroup = (int *)malloc(r*sizeof(int));

    for (i = 0; i < r; ++i)
       regroup[i]=i;
    for (j = k-1; j >= 0; --j)
    {
       for(i = 0; i < na; ++i)
         sxc[i]=0;
       for (i = 0; i < r; ++i)
       {
                cc[i]=sx[regroup[i]+comb[j]*r];
		sxc[ cc[i] ]++;
       }
       bc[0]=0;
       for (i = 1;i < na; ++i)
           bc[i]=bc[i-1]+sxc[i-1];
       for (i = 0; i < r; ++i)
            sxl[bc[ cc[i] ]++] = regroup[i];
       for (i = 0; i < r; ++i)
            regroup[i] = sxl[i];
    }
    free(sxl); free(bc); free(sxc); free(cc);

    return regroup;
}


int *cntsrtna_gapped(int *sx, int g, int k, int r, int na, Combination *combinations)
{
    /* This implements implicit sorting */
    int *sxc, *bc, *sxl, *cc, *regroup;
    int i, j, ind, rk, indc;
    int ncomb;
    
    ncomb = (int) (*combinations).num_comb;
    rk = r * ncomb; 
    sxc = (int *)malloc(na*sizeof(int));
    bc  = (int *)malloc(na*sizeof(int));
    sxl = (int *)malloc(rk*sizeof(int));
    cc  = (int *)malloc(rk*sizeof(int));
    regroup = (int *)malloc(rk*sizeof(int));
    for (i = 0; i < rk; ++i)
      regroup[i] = i;

    for (j = k-1; j >= 0; --j)
    {
      for(i = 0; i < na; ++i)
         sxc[i]=0;
      for (i = 0; i < rk; ++i)
      {
        indc = (int)((double)regroup[i]/(double)r);
        ind = regroup[i] - indc * r;
        cc[i] = sx[ind + combinations[0].comb[indc+j*ncomb]*r];
        sxc[ cc[i] ]++; 
      }
       bc[0]=0;
       for (i = 1;i < na; ++i)
           bc[i]=bc[i-1]+sxc[i-1];
       for (i = 0; i < rk; ++i)
            sxl[bc[ cc[i] ]++] = regroup[i];
       for (i = 0; i < rk; ++i)
            regroup[i] = sxl[i];
    }
    free(sxl); free(bc); free(sxc); free(cc);

    return regroup;
}

void countAndUpdate(int *outK, int *sx, int *g, int k, int r, int nStr)
{
   char same;
   int i, j; 
   int cu;
   long int ucnt;
   long int startInd, endInd, j1;
   int *curfeat, *ucnts, *updind;

   curfeat = (int *)malloc(k*sizeof(int));
   ucnts= (int *)malloc(nStr*sizeof(int));
   updind = (int *)malloc(nStr*sizeof(int));
   i = 0;
   ucnt = 0;
   while (i<r)
   {
     for (j = 0; j < k; ++j)
	curfeat[j]=sx[i+j*r];
     same=1;
     for (j = 0;j < k; ++j)
	if (curfeat[j]!=sx[i+j*r])
	{
	   same=0;
	   break;
	}
     startInd=i;
     while (same && i<r)
     {
	i++;
	if (i >= r) break;
        same = 1;
	for (j = 0; j < k; ++j)
	   if (curfeat[j]!=sx[i+j*r])
	   {
	   	same=0;
	   	break;
	   }
     }
     endInd= (i<r) ? (i-1):(r-1);
     ucnt++;
     if ((long int)endInd-startInd+1>2*nStr)	
     {
	for (j = 0; j < nStr; ++j) ucnts[j]=0;
        for (j = startInd;j <= endInd; ++j)  ucnts[g[j]]++;
	cu=0;
	for (j=0;j<nStr;j++)
	{
	   if (ucnts[j]>0)
	   {
	     updind[cu] = j;
             cu++;
	   }
	}
	for (j=0;j<cu;j++)
	  for (j1=0;j1<cu;j1++)
	    outK[updind[j]+updind[j1]*nStr]+=ucnts[updind[j]]*ucnts[updind[j1]];
     }
     else
     {
        for (j = startInd;j <= endInd; ++j)
          for (j1 = startInd;j1 <= endInd; ++j1)
            outK[ g[j]+nStr*g[j1] ]++;
     }   
  }
  free(updind);
  free(ucnts);
  free(curfeat);
}


void countAndUpdate_vector(int *outK, int *sx, int *g, int k, int r, int nStr, int nq)
{
   char same;
   int i, j; 
   int cu;
   long int ucnt;
   long int startInd, endInd, j1;
   int *curfeat, *ucnts, *updind;

   curfeat = (int *)malloc(k*sizeof(int));
   ucnts= (int *)malloc(nStr*sizeof(int));
   updind = (int *)malloc(nStr*sizeof(int));
   i = 0;
   ucnt = 0;
   while (i<r)
   {
     for (j = 0; j < k; ++j)
	curfeat[j]=sx[i+j*r];
     same=1;
     for (j = 0;j < k; ++j)
	if (curfeat[j]!=sx[i+j*r])
	{
	   same=0;
	   break;
	}
     startInd=i;
     while (same && i<r)
     {
	i++;
	if (i >= r) break;
        same = 1;
	for (j = 0; j < k; ++j)
	   if (curfeat[j]!=sx[i+j*r])
	   {
	   	same=0;
	   	break;
	   }
     }
     endInd= (i<r) ? (i-1):(r-1);
     ucnt++;
     if ((long int)endInd-startInd+1>2*nStr)	
     {
	for (j = 0; j < nStr; ++j) ucnts[j]=0;
        for (j = startInd;j <= endInd; ++j)  ucnts[g[j]]++;
	cu=0;
	for (j=0;j<nStr;j++)
	{
	   if (ucnts[j]>0)
	   {
	     updind[cu] = j;
             cu++;
	   }
	}
	for (j = 0; j < nq; ++j)
          if (ucnts[j]>0)
   	    for (j1 = 0; j1 < cu; ++j1)
	     outK[j+updind[j1]*nStr]+=ucnts[j]*ucnts[updind[j1]];
     }
     else
     {
        for (j = startInd;j <= endInd; ++j)
          if (g[j] < nq)
            for (j1 = startInd;j1 <= endInd; ++j1)
              outK[ g[j]+nStr*g[j1] ]++;
     }   
  }
  free(updind);
  free(ucnts);
  free(curfeat);
}

void countAndUpdate_comb_vector(int *outK, int *sx, int *g, int k, int r, int nStr, int *regroup, int *comb, int nq)
{
   char same;
   int i, j; 
   int cu;
   long int ucnt;
   long int startInd, endInd, j1, ind1;
   int *curfeat, *ucnts, *updind, *g1;

   curfeat = (int *)malloc(k*sizeof(int));
   ucnts= (int *)malloc(nStr*sizeof(int));
   updind = (int *)malloc(nStr*sizeof(int));
   

   g1 = (int *)malloc(r*sizeof(int));
   for (i = 0; i < r; ++i)
     g1[i] = g[regroup[i]];

   i = 0;
   ucnt = 0;
   while (i<r)
   {
     for (j = 0; j < k; ++j)
	curfeat[j]=sx[regroup[i]+comb[j]*r];
     same=1;
     for (j = 0;j < k; ++j)
	if (curfeat[j]!=sx[regroup[i]+comb[j]*r])
	{
	   same=0;
	   break;
	}
     startInd=i;
     while (same && i<r)
     {
	i++;
	if (i >= r) break;
        same = 1;
	for (j = 0; j < k; ++j)
	   if (curfeat[j]!=sx[regroup[i]+comb[j]*r])
	   {
	   	same=0;
	   	break;
	   }
     }
     endInd= (i<r) ? (i-1):(r-1);
     ucnt++;
     if ((long int)endInd-startInd+1>2*nStr)	
     {
	for (j = 0; j < nStr; ++j) ucnts[j]=0;
        for (j = startInd;j <= endInd; ++j)  ucnts[g1[j]]++;
	cu=0;
	for (j=0;j<nStr;j++)
	{
	   if (ucnts[j]>0)
	   {
	     updind[cu] = j;
             cu++;
	   }
	}
	for (j = 0; j < nq; ++j)
          if (ucnts[j]>0)
  	    for (j1 = 0; j1 < cu; ++j1)
	      outK[j+updind[j1]*nStr]+=ucnts[j]*ucnts[updind[j1]];
     }
     else
     {
        for (j = startInd;j <= endInd; ++j)
        {
          ind1 = regroup[j];
          if (g1[j] < nq)
          for (j1 = startInd;j1 <= endInd; ++j1)
            outK[ g1[j]+nStr*g1[j1] ]++;
        }
     }   
  }
  free(g1);
  free(updind);
  free(ucnts);
  free(curfeat);
}


void countAndUpdate_comb(int *outK, int *sx, int *g, int k, int r, int nStr, int *regroup, int *comb)
{
   char same;
   int i, j; 
   int cu;
   long int ucnt;
   long int startInd, endInd, j1;
   int ind1;
   int *curfeat, *ucnts, *updind, *g1;

   curfeat = (int *)malloc(k*sizeof(int));
   ucnts= (int *)malloc(nStr*sizeof(int));
   updind = (int *)malloc(nStr*sizeof(int));
   

   g1 = (int *)malloc(r*sizeof(int));
   for (i = 0; i < r; ++i)
     g1[i] = g[regroup[i]];

   i = 0;
   ucnt = 0;
   while (i<r)
   {
     for (j = 0; j < k; ++j)
	curfeat[j]=sx[regroup[i]+comb[j]*r];
     same=1;
     for (j = 0;j < k; ++j)
	if (curfeat[j]!=sx[regroup[i]+comb[j]*r])
	{
	   same=0;
	   break;
	}
     startInd=i;
     while (same && i<r)
     {
	i++;
	if (i >= r) break;
        same = 1;
	for (j = 0; j < k; ++j)
	   if (curfeat[j]!=sx[regroup[i]+comb[j]*r])
	   {
	   	same=0;
	   	break;
	   }
     }
     endInd= (i<r) ? (i-1):(r-1);
     ucnt++;
     if ((long int)endInd-startInd+1>2*nStr)	
     {
	for (j = 0; j < nStr; ++j) ucnts[j]=0;
        for (j = startInd;j <= endInd; ++j)  ucnts[g1[j]]++;
	cu=0;
	for (j=0;j<nStr;j++)
	{
	   if (ucnts[j]>0)
	   {
	     updind[cu] = j;
             cu++;
	   }
	}
	for (j=0;j<cu;j++)
	  for (j1=0;j1<cu;j1++)
	    outK[updind[j]+updind[j1]*nStr]+=ucnts[updind[j]]*ucnts[updind[j1]];
     }
     else
     {
        for (j = startInd;j <= endInd; ++j)
        {
          ind1 = regroup[j];
          for (j1 = startInd;j1 <= endInd; ++j1)
            outK[ g1[j]+nStr*g1[j1] ]++;
        }
     }   
  }
  free(g1);
  free(updind);
  free(ucnts);
  free(curfeat);
}


void countAndUpdateN2(int *outK, int *sx, int *g, int k, int r, int nStr)
{
   /* Strictly uses O(N^2) update */
   char same;
   int i, j; 
   int cu;
   long int ucnt;
   long int startInd, endInd, j1;
   int *curfeat, *ucnts, *updind;

   curfeat = (int *)malloc(k*sizeof(int));
   ucnts= (int *)malloc(nStr*sizeof(int));
   updind = (int *)malloc(nStr*sizeof(int));
 
   i = 0;
   ucnt = 0;
   while (i<r)
   {
     for (j = 0; j < k; ++j)
	curfeat[j]=sx[i+j*r];
     same=1;
     for (j = 0;j < k; ++j)
	if (curfeat[j]!=sx[i+j*r])
	{
	   same=0;
	   break;
	}
     startInd=i;
     while (same && i<r)
     {
	i++;
	if (i >= r) break;
        same = 1;
	for (j = 0; j < k; ++j)
	   if (curfeat[j]!=sx[i+j*r])
	   {
	   	same=0;
	   	break;
	   }
     }
     endInd= (i<r) ? (i-1):(r-1);
     ucnt++;
   /*  if ((long int)endInd-startInd+1>2*nStr)	
     {*/
	for (j = 0; j < nStr; ++j) ucnts[j]=0;
        for (j = startInd;j <= endInd; ++j)  ucnts[g[j]]++;
	cu=0;
	for (j=0;j<nStr;j++)
	{
	   if (ucnts[j]>0)
	   {
	     updind[cu] = j;
             cu++;
	   }
	}
	for (j=0;j<cu;j++)
	  for (j1=0;j1<cu;j1++)
	    outK[updind[j]+updind[j1]*nStr]+=ucnts[updind[j]]*ucnts[updind[j1]];
    /* }
     else
     {
        for (j = startInd;j <= endInd; ++j)
          for (j1 = startInd;j1 <= endInd; ++j1)
            outK[ g[j]+nStr*g[j1] ]++;
     }*/   
  }
  free(updind);
  free(ucnts);
  free(curfeat);
}


void countAndUpdate_double(double *outK, int *sx, int *g, int k, int r, int nStr)
{
   char same;
   int i, j; 
   int cu;
   long int ucnt;
   long int startInd, endInd, j1;
   int *curfeat, *ucnts, *updind;

   curfeat = (int *)malloc(k*sizeof(int));
   ucnts= (int *)malloc(nStr*sizeof(int));
   updind = (int *)malloc(nStr*sizeof(int));
 
   i = 0;
   ucnt = 0;
   while (i<r)
   {
     for (j = 0; j < k; ++j)
	curfeat[j]=sx[i+j*r];
     same=1;
     for (j = 0;j < k; ++j)
	if (curfeat[j]!=sx[i+j*r])
	{
	   same=0;
	   break;
	}
     startInd=i;
     while (same && i<r)
     {
	i++;
	if (i >= r) break;
        same = 1;
	for (j = 0; j < k; ++j)
	   if (curfeat[j]!=sx[i+j*r])
	   {
	   	same=0;
	   	break;
	   }
     }
     endInd= (i<r) ? (i-1):(r-1);
     ucnt++;
     /*if ((long int)endInd-startInd+1>2*nStr)	
     {*/
	for (j = 0; j < nStr; ++j) ucnts[j]=0;
        for (j = startInd;j <= endInd; ++j)  ucnts[g[j]]++;
	cu=0;
	for (j=0;j<nStr;j++)
	{
	   if (ucnts[j]>0)
	   {
	     updind[cu] = j;
             cu++;
	   }
	}
	for (j=0;j<cu;j++)
	  for (j1=0;j1<cu;j1++)
	    outK[updind[j]+updind[j1]*nStr]+=(double)ucnts[updind[j]]*ucnts[updind[j1]];
   /*}
     else
     {
        for (j = startInd;j <= endInd; ++j)
          for (j1 = startInd;j1 <= endInd; ++j1)
            outK[ g[j]+nStr*g[j1] ]++;
     } */ 
  }
  free(updind);
  free(ucnts);
  free(curfeat);
}


void countAndUpdate_double_gapped(double *outK, int *sx, int *g, int *regroup, int k, int r, int nStr, Combination *combinations)
{
   /* This uses implicit counting for gapped kernel */
   char same;
   int i, j; 
   int cu;
   long int ucnt;
   long int startInd, endInd, j1;
   int indc, ind;
   int *curfeat, *ucnts, *updind;
   int rk,ncomb;

   ncomb = (int)combinations[0].num_comb;
   curfeat = (int *)malloc(k*sizeof(int));
   ucnts= (int *)malloc(nStr*sizeof(int));
   updind = (int *)malloc(nStr*sizeof(int));
 
   i = 0;
   ucnt = 0;
   rk = r * ncomb;
   printf("k=%d,nStr=%d,r=%d,rk=%d\n",k,nStr,r,rk);
   while (i < rk)
   {
     for (j = 0; j < k; ++j)
     {
        indc = (int)((double)regroup[i]/(double)r);
        ind = regroup[i] - indc * r;
	curfeat[j]=sx[ind + combinations[0].comb[indc + j*ncomb]*r];
     }
     same = 1;
     for (j = 0;j < k; ++j)
     {
        indc = (int)((double)regroup[i]/(double)r);
        ind = regroup[i] - indc * r;
	if (curfeat[j]!=sx[ind + combinations[0].comb[indc + j*ncomb]*r]) 
	{
	   same=0;
	   break;
	}
     }
     startInd=i;
     while (same && i<rk)
     {
	i++;
       	if (i >= rk) break;
        same = 1;
	for (j = 0; j < k; ++j)
        {
        indc = (int)((double)regroup[i]/(double)r);
        ind = regroup[i] - indc * r;
	if (curfeat[j]!=sx[ind + combinations[0].comb[indc + j*ncomb]*r]) 
	   {
	   	same=0;
	   	break;
	   }
        }
     }
     endInd= (i<rk) ? (i-1):(rk-1);
     ucnt++;
     /*if ((long int)endInd-startInd+1>2*nStr)	
     {*/
	for (j = 0; j < nStr; ++j) ucnts[j]=0;
        for (j = startInd;j <= endInd; ++j)
        {
           indc = (int)((double)regroup[j]/(double)r);
           ind = regroup[j] - indc * r;
           ucnts[g[ind]]++;
        }
	cu=0;
	for (j=0;j<nStr;j++)
	{
	   if (ucnts[j]>0)
	   {
	     updind[cu] = j;
             cu++;
	   }
	}
	for (j=0;j<cu;j++)
	  for (j1=0;j1<cu;j1++)
	    outK[updind[j]+updind[j1]*nStr]+=(double)ucnts[updind[j]]*ucnts[updind[j1]];
/*   }
     else
     {
        for (j = startInd;j <= endInd; ++j)
        {
          indc = (int)((double)regroup[j]/(double)r);
          ind = regroup[j] - indc * r;
          for (j1 = startInd;j1 <= endInd; ++j1)
          {
       
            indc = (int)((double)regroup[j1]/(double)r);
            ind1 = regroup[j1] - indc * r;
            outK[ g[ind]+nStr*g[ind1] ]++;
          }
        }
     }  */
  }
  free(updind);
  free(ucnts);
  free(curfeat);
}


void countAndUpdate_double_gapped_vector(double *outK, int *sx, int *g, int *regroup, int k, int r, int nStr, Combination *combinations, int nq)
{
   /* This uses implicit counting for gapped kernel */
   char same;
   int i, j; 
   int cu;
   long int ucnt;
   long int startInd, endInd, j1;
   int indc, ind;
   int *curfeat, *ucnts, *updind;
   int rk,ncomb;

   ncomb = (int)combinations[0].num_comb;
   curfeat = (int *)malloc(k*sizeof(int));
   ucnts= (int *)malloc(nStr*sizeof(int));
   updind = (int *)malloc(nStr*sizeof(int));
 
   i = 0;
   ucnt = 0;
   rk = r * ncomb;
   printf("k=%d,nStr=%d,r=%d,rk=%d\n",k,nStr,r,rk);
   while (i < rk)
   {
     for (j = 0; j < k; ++j)
     {
        indc = (int)((double)regroup[i]/(double)r);
        ind = regroup[i] - indc * r;
	curfeat[j]=sx[ind + combinations[0].comb[indc + j*ncomb]*r];
     }
     same = 1;
     for (j = 0;j < k; ++j)
     {
        indc = (int)((double)regroup[i]/(double)r);
        ind = regroup[i] - indc * r;
	if (curfeat[j]!=sx[ind + combinations[0].comb[indc + j*ncomb]*r]) 
	{
	   same=0;
	   break;
	}
     }
     startInd=i;
     while (same && i<rk)
     {
	i++;
       	if (i >= rk) break;
        same = 1;
	for (j = 0; j < k; ++j)
        {
        indc = (int)((double)regroup[i]/(double)r);
        ind = regroup[i] - indc * r;
	if (curfeat[j]!=sx[ind + combinations[0].comb[indc + j*ncomb]*r]) 
	   {
	   	same=0;
	   	break;
	   }
        }
     }
     endInd= (i<rk) ? (i-1):(rk-1);
     ucnt++;
     /*if ((long int)endInd-startInd+1>2*nStr)	
     {*/
	for (j = 0; j < nStr; ++j) ucnts[j]=0;
        for (j = startInd;j <= endInd; ++j)
        {
           indc = (int)((double)regroup[j]/(double)r);
           ind = regroup[j] - indc * r;
           ucnts[g[ind]]++;
        }
	cu=0;
	for (j=0;j<nStr;j++)
	{
	   if (ucnts[j]>0)
	   {
	     updind[cu] = j;
             cu++;
	   }
	}
	for (j = 0; j < nq; ++j)
          if (ucnts[j]>0)
	  for (j1 = 0; j1 < cu; ++j1)
	    outK[j+updind[j1]*nStr]+=(double)ucnts[j]*ucnts[updind[j1]];
/*   }
     else
     {
        for (j = startInd;j <= endInd; ++j)
        {
          indc = (int)((double)regroup[j]/(double)r);
          ind = regroup[j] - indc * r;
          for (j1 = startInd;j1 <= endInd; ++j1)
          {
       
            indc = (int)((double)regroup[j1]/(double)r);
            ind1 = regroup[j1] - indc * r;
            outK[ g[ind]+nStr*g[ind1] ]++;
          }
        }
     }  */
  }
  free(updind);
  free(ucnts);
  free(curfeat);
}

double nchoosek(int n, int k)
{
  int i;
  double *nums, *dens;
  double prod;
  
  prod = 1;
  if (k > n/2) k = n-k;
  if (k==0) return 1;
  if (k==1) return n;
  if (k > 1)
  { 
    nums = malloc(k*sizeof(double));
    dens = malloc(k*sizeof(double));
    for (i = 0; i < k; ++i)
    {
       nums[i] = n-k+1+i;
       dens[i] = i+1;
    }  
    prod = 1;
    for (i = 0; i < k; ++i)
    {
      nums[i] = nums[i] / dens[i];
      prod *=  nums[i];
    }
    free(nums);
    free(dens);
  }
  return prod;
}

Combination *getCombinations(int n, int k)
{
    Combination *combinations;
    int *cnt_comb;
    int *elems, *pos;
    int depth, margin;
    int ii;
   
    combinations = (Combination *)malloc(sizeof(Combination));
    cnt_comb = (int *)malloc(sizeof(int));
    cnt_comb[0] = 0;
    combinations[0].n = n;
    combinations[0].k = k;
    combinations[0].num_comb = 1;
    if (n > 0 && k > 0)
    {
      combinations[0].num_comb = nchoosek((double)n,(double)k);
      combinations[0].comb = (int *)malloc(combinations[0].num_comb*combinations[0].k*sizeof(int)); 
      elems = (int *)malloc((n+1)*sizeof(int));
      pos   = (int *)malloc((n+1)*sizeof(int));
      depth = 0;
      margin = 0;
      for (ii = 0; ii < n; ++ii)
         elems[ii] = ii;
      traverseCombinations(elems, n, k, pos, depth, margin, cnt_comb, combinations[0].comb,combinations[0].num_comb);
      free(elems);
      free(pos);

    }

      return combinations;
}

void traverseCombinations(const int *elems, int n, int k, int *pos, int depth, int margin, int *cnt_comb, int *out, int num_comb)
{
 int j;
 if (depth >= k)
 {
    for (j = 0; j < k; ++j)
       out[cnt_comb[0] + j*num_comb] = elems[pos[j]];
    cnt_comb[0]++;    
    return; 
 }
 for (j = margin; j < n; ++j)
 {
    pos[depth]=j;
    traverseCombinations(elems, n, k, pos, depth + 1, j + 1, cnt_comb, out, num_comb);    
 }
}

int saveKernelInt(int *K, int n, int m, const char *format, char *filename)
{
  int i, j;
  FILE *kernelfile;
  kernelfile = fopen(filename, "w");
  if ( kernelfile )
  {
    for (i = 0; i < n; ++i)
    {
       for (j = 0; j < m-1; ++j)
         fprintf(kernelfile, format, K[i+j*n]);
       fprintf(kernelfile, "%d", K[i+(m-1)*n]);
       fprintf(kernelfile,"\n");
    }   
    fclose(kernelfile);
  }
  else
  {
    perror("Cannot write kernel file\n");
    exit(1);
  }

  return 0;
} 

int saveKernelDouble(double *K, int n, int m, const char *format, char *filename)
{
  int i, j;
  FILE *kernelfile;
  kernelfile = fopen(filename, "w");
  if ( kernelfile )
  {
    for (i = 0; i < n; ++i)
    {
       for (j = 0; j < m-1; ++j)
         fprintf(kernelfile, format, K[i+j*n]);
       fprintf(kernelfile, "%.4f", K[i+(m-1)*n]);
       fprintf(kernelfile,"\n");
    }   
    fclose(kernelfile);
  }
  else
  {
    perror("Cannot write kernel file\n");
    exit(1);
  }
  return 0;
} 
