#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "shared.h"

#define ARG_REQUIRED 3

Features *extractTriple(int **S, int *len, int nStr, int d1, int d2)
{
  int i, j;
  int n, nfeat;
  int *group;
  int *features;
  int *s;
  int c;
  Features *F;

  nfeat = 0;
  for (i = 0; i < nStr; ++i)
    nfeat += len[i] - d1 -d2;
  group = (int *)malloc(nfeat*sizeof(int));
  features = (int *)malloc(nfeat*3*sizeof(int *));
  c = 0;
  for (i = 0; i < nStr; ++i)
  {
    n = len[i] - d1 - d2;
    s = S[i];
    for (j = 0; j < n; ++j)
    {
      features[c] = s[j];
      features[c+nfeat] = s[j+d1];
      features[c+2*nfeat] = s[j+d1+d2];
      group[c] = i;
      c++;
    }
  }
  if (nfeat!=c)
    printf("Something is wrong...\n");
  F = (Features *)malloc(sizeof(Features));
  (*F).features = features;
  (*F).group = group;
  (*F).n = nfeat;
  return F;
}


int help()
{
  printf("Usage: tripleKernel <Sequence-file> <d> <Alphabet-size>\n");
  printf("\t Sequence-file : file with sequence data\n");
  printf("\t d : kernel parameter (max. distance between samples), d>0\n");
  printf("\t Alphabet-size: size of the sequence alphabet, >0\n");
  printf("\t IMPORTANT: sequence elements must be\n\tin the range [0,AlphabetSize - 1].\n");
  return 1;
}

char isVerbose;
int *tripleKernel(int **S, int d, int nStr, int na, int *len);

int main (int argc, char **argv)
{
  
  char *filename;
  int d, na, nStr;
  int *K;
  int **S, *len;
  char *kernelfilename;
  
  isVerbose = 0;
  if (argc != ARG_REQUIRED+1)
  {
    return help();
  }

  filename = argv[1];
  d = atoi(argv[2]);
  na = atoi(argv[3]);
  if (d <= 0 || na <=0)
  {
    printf("Error in input parameters\n");
    return help();
  }
  len = (int *)malloc(MAXNSTR*sizeof(int));
  printf("Reading %s\n",filename);
  S = loadStrings(filename, len, &nStr);
  printf("Read %d strings\n", nStr);
  K = tripleKernel(S, d, nStr, na, len);
  kernelfilename = "kernel.txt";
  saveKernelInt(K,nStr,nStr,"%d ",kernelfilename);
  free(K);
  return 0;
}

int *tripleKernel(int **S, int d, int nStr, int na, int *len)
{
  int *K, *features_srt, *group_srt;
  int *sortIdx;
  int i, d1, d2, nfeat, maxIdx;
  Features *features;

  K = (int *)malloc(nStr*nStr*sizeof(int));
  for (i=0; i<nStr*nStr; ++i)
    K[i]=0;
  printf("Computing triple(1,%d) kernel\n",d);
  for (d1 = 1; d1 <= d; ++d1)
  {
    for (d2 = 1; d2 <= d; ++d2)
    {
      if (isVerbose)
      {
        printf("Extracting features...");
        fflush( stdout );
      }
      features = extractTriple(S,len,nStr,d1,d2);
      nfeat = (*features).n;
      printf("(%d,%d): %d features\n", d1, d2, nfeat);
      maxIdx = 0;
      for (i = 0; i < nfeat*3; ++i)
        maxIdx = maxIdx > (*features).features[i] ? maxIdx : (*features).features[i];
      if (na < maxIdx+1)
      {
        printf("WARNING: Sequence elements are outside the specified range [0,%d]\n",na);
        printf("\tUsing [0,%d] instead\n", maxIdx);
        na = maxIdx+1;
      }
      if (isVerbose)
      {
        printf("done.\n");
        printf("Sorting...");
        fflush( stdout );
      }
      sortIdx = cntsrtna((*features).features,3,(*features).n,na);
      if (isVerbose)
      {
        printf("done.\n");
      }
      features_srt = (int *)malloc(nfeat*3*sizeof(int *));
      group_srt = (int *)malloc(nfeat*sizeof(int));
      for (i = 0; i < nfeat; ++i)
      {
        features_srt[i]=(*features).features[sortIdx[i]];
        features_srt[i+nfeat]=(*features).features[sortIdx[i]+nfeat];
        features_srt[i+2*nfeat]=(*features).features[sortIdx[i]+2*nfeat];
        group_srt[i] = (*features).group[sortIdx[i]];
      }
      free(sortIdx);
      free((*features).features);
      free((*features).group);
      free(features);
      if (isVerbose)
      {
        printf("Counting...");
        fflush( stdout );
      }
      countAndUpdate(K,features_srt,group_srt,3,nfeat,nStr);
      free(features_srt);
      free(group_srt);
      if (isVerbose)
      {
        printf("done.\n");
      }
    }
  }
  

  return K;
}

